
# script to analyze and average protein torsion angles
#
# usage:
#   xplor_nih -py -quiet -notrace -psf=[psf file] tang.py [pdb files] >average.info
#
# where [pdb files] is one or more structure files
# and information averaged over all the structures is send to stdout
#
# for each structure, a printout of all the torsion angles is saved to 
# a file named structN.angles
#

from sys import argv
from selectTools import minResid, maxResid
from simulationTools import StructureLoop
from dihedral import Dihedral
from pdbTool import PDBTool
import protocol

psfRead=0
for arg in argv:
    if arg.startswith('-psf='):
        psf=arg.split('=')[1]
        protocol.initStruct(psf)
        psfRead=1
        pass
    pass

#if not psfRead: protocol.initStruct("/home/schwitrs/clore/gb3/gb3.psf")

def getAngles(resType,resid):
    angles = [("phi", ("name C  and resid %d" % (resid-1,),
                       "name N  and resid %d" % (resid,),
                       "name CA and resid %d" % (resid,),
                       "name C  and resid %d" % (resid,) )),
              ("psi",("name N  and resid %d" % (resid,),
                      "name CA and resid %d" % (resid,),
                      "name C  and resid %d" % (resid,),
                      "name N  and resid %d" % (resid+1,))),
              ("omega",("name CA and resid %d" % (resid,),
                        "name C  and resid %d" % (resid,),
                        "name N  and resid %d" % (resid+1,),
                        "name CA and resid %d" % (resid+1,)))]
    try:
        for (name,atoms) in scTorsionAtoms[resType]:
            angles.append((name,("name %s and resid %d" % (atoms[0],resid),
                                 "name %s and resid %d" % (atoms[1],resid),
                                 "name %s and resid %d" % (atoms[2],resid),
                                 "name %s and resid %d" % (atoms[3],resid))))
            pass
        pass
    except KeyError:
        print "unsupported residue type:", resType
        pass
        
    return angles

scTorsionAtoms = {}
scTorsionAtoms['GLY'] = []
scTorsionAtoms['ALA'] = []
scTorsionAtoms['SER'] = [('CHI1',('N', 'CA', 'CB', 'OG')),
                         ('CHI2',('CA','CB', 'OG', 'HG'))]
scTorsionAtoms['THR'] = [('CHI1' , ('N' , 'CA', 'CB', 'OG1')),
                         ('CHI21', ('CA', 'CB', 'OG1', 'HG1'))]
scTorsionAtoms['LYS'] = [('CHI1', ('N',  'CA', 'CB', 'CG' )),
                         ('CHI2', ('CA', 'CB', 'CG', 'CD' )),
                         ('CHI3', ('CB', 'CG', 'CD', 'CE' )),
                         ('CHI4', ('CG', 'CD', 'CE', 'NZ' ))]
scTorsionAtoms['CYS'] =[('CHI1', ('N',  'CA', 'CB', 'SG')),
                        ('CHI2', ('CA', 'CB', 'SG', 'HG'))]
scTorsionAtoms['MET'] = [('CHI1', ('N',  'CA', 'CB', 'CG' )),
                         ('CHI2', ('CA', 'CB', 'CG', 'SD' )),
                         ('CHI3', ('CB', 'CG', 'SD', 'CE' ))]
scTorsionAtoms['VAL'] =[('CHI1',  ('N',  'CA', 'CB',  'CG1' ))]
scTorsionAtoms['ILE'] = [('CHI1',  ('N',  'CA',  'CB',  'CG1')),
                         ('CHI21', ('CA', 'CB',  'CG1', 'CD1'))]
scTorsionAtoms['LEU'] = [('CHI1',  ('N',  'CA',  'CB',  'CG'   )),
                         ('CHI2',  ('CA', 'CB',  'CG',  'CD1'  ))]
scTorsionAtoms['ASP'] = [('CHI1',  ('N',  'CA', 'CB',  'CG'  )),
                         ('CHI2',  ('CA', 'CB', 'CG',  'OD1' ))]
scTorsionAtoms['ASN'] = [('CHI1',  ('N',  'CA', 'CB',  'CG'  )),
                         ('CHI2',  ('CA', 'CB', 'CG',  'OD1' ))]
scTorsionAtoms['GLU'] = [('CHI1',  ('N',  'CA', 'CB',  'CG'  )),
                         ('CHI2',  ('CA', 'CB', 'CG',  'CD'  )),
                         ('CHI3',  ('CB', 'CG', 'CD',  'OE1' ))]
scTorsionAtoms['GLN'] = [('CHI1',  ('N',  'CA', 'CB',  'CG'  )),
                         ('CHI2',  ('CA', 'CB', 'CG',  'CD'  )),
                         ('CHI3',  ('CB', 'CG', 'CD',  'OE1' ))]
scTorsionAtoms['ARG'] = [('CHI1',  ('N',  'CA', 'CB',  'CG'  )),
                         ('CHI2',  ('CA', 'CB', 'CG',  'CD'  )),
                         ('CHI3',  ('CB', 'CG', 'CD',  'NE'  )),
                         ('CHI4',  ('CG', 'CD', 'NE',  'CZ'  ))]
scTorsionAtoms['PRO'] = []
scTorsionAtoms['HIS'] = [('CHI1',  ('N',  'CA', 'CB', 'CG'  )),
                         ('CHI2',  ('CA', 'CB', 'CG', 'ND1' ))]
scTorsionAtoms['PHE'] = [('CHI1',  ('N',  'CA', 'CB', 'CG' )),
                         ('CHI2',  ('CA', 'CB', 'CG', 'CD1'))]
scTorsionAtoms['TYR'] = [('CHI1',   ('N',   'CA', 'CB', 'CG' )),
                         ('CHI2',   ('CA',  'CB', 'CG', 'CD1')),
                         ('CHI6',   ('CE1', 'CZ', 'OH', 'HH' ))]
scTorsionAtoms['TRP'] = [('CHI1',  ('N',  'CA', 'CB', 'CG' )),
                         ('CHI2',  ('CA', 'CB', 'CG', 'CD1'))]

binData={}
#binData['CHI1'] = (('180',180), ('+60',60), ('-60',-60))
#binData['CHI2'] = (('180',180), ('+60',60), ('-60',-60))
#binData['CHI3'] = (('180',180), ('+60',60), ('-60',-60))
#binData['CHI4'] = (('180',180), ('+60',60), ('-60',-60))
#binData['CHI5'] = (('180',180), ('+60',60), ('-60',-60))

binData['PHE_CHI2'] = (('+90',90), ('-90',-90))
binData['TYR_CHI2'] = (('+90',90), ('-90',-90))
binData['TRP_CHI2'] = (('+90',90), ('-90',-90))
binData['HIS_CHI2'] = (('+90',90), ('-90',-90))
binData['ASP_CHI2'] = (('0',0), ('180',180))
binData['ASN_CHI2'] = (('0',0), ('180',180))
binData['GLU_CHI3'] = (('0',0), ('180',180))
binData['GLN_CHI3'] = (('0',0), ('180',180))

def getBranch(theta1,theta2):
    """return the branch number n such that n*2*PI+theta1-theta2 is
    minimal"""
    return round((theta2-theta1)/360)

def getBin(resType,taname,val):
    closest = ('single',0,val)
    bins=[]
    if taname.startswith('CHI'):
        bins = (('180',180), ('+60',60), ('-60',-60))
        pass
    try:
        bins = binData[resType+'_'+taname]
    except KeyError:
        pass
    
    if bins:
        closest = ('',361,361)
        for (name,mean) in bins:
            val += 360*getBranch(val,mean)
            dist=abs(val-mean)
            if dist<closest[1]:
                closest = (name,dist,val)
                pass
            #print 'branch',mean,val,getBranch(val,mean), closest
            pass
        pass
        
    return closest
        

    

angleBins={}
for resid in range(minResid(),maxResid()+1):
    angleBins[resid] = {}
    pass
def binAngle(resid,resType,taName,val):
    global angleBins
    (binName,binMean,branchVal) = getBin(resType,taName,val)
    while 1:
        try:
            tabin = angleBins[resid][taName]
        except KeyError:
            angleBins[resid][taName] = {}
            tabin = angleBins[resid][taName]
            pass
        try:
            tabin[binName].append(branchVal)
            break
        except KeyError:
            tabin[binName] = []
            pass
        pass
    return
            
        


def binAngles(filename):
    pdbFile = PDBTool(filename)
    pdbFile.read()
    afile = open(pdbFile.filename()+".angles",'w')

    #FIX: this assumes only a single segid
    for resid in range(minResid(),maxResid()+1):
        resType = AtomSel("resid %d"%resid)[0].residueName()
        print >> afile, "%3s %4d" % (resType, resid),
        for (aname,atomSel) in getAngles( resType , resid):
            val=0
            try:
                val=Dihedral(*atomSel).value()
                print >> afile,  "%7.2f" % val,
                binAngle(resid,resType,aname,val)
            except IndexError:
                print >> afile, "[%5s]"%aname,
                pass
            pass
        print >> afile
        pass
    return

scriptPos = len(argv)
for cnt in range(len(argv)):
    if argv[cnt][0] != '-':
        scriptPos = cnt
        break
    pass

for filename in  argv[scriptPos+1:]:
    binAngles(filename)
    pass


def average(l):
    ret=0.
    if not l: return ret
    
    for el in l:
        ret += el
        pass
    ret /= len(l)
    return ret

def deviation(l):
    from math import sqrt
    ret=0.
    if not l: return ret

    ave = average(l)

    for el in l:
        ret += (el-ave)**2
        pass
    ret /= len(l)
    return sqrt(ret)


print "resid            bin    cnt     ave     dev"
for resid in range(minResid(),maxResid()+1):
    atom = AtomSel("tag and resid %d"%resid)[0]
    resType = atom.residueName()
    angles=angleBins[resid].keys()
    angles.sort()
    for angle in angles:
        bins = angleBins[resid][angle].keys()
        bins.sort()
        for bin in bins:
            print "%5d %3s %5s %6s" % (resid, resType, angle, bin),
            print "%4d" % len(angleBins[resid][angle][bin]),
            print "%7.2f" % average(angleBins[resid][angle][bin]),
            print "%7.2f" % deviation(angleBins[resid][angle][bin])
            pass
        pass
    pass
            
